#include "miniaudioengine.h"

MiniAudioEngine::MiniAudioEngine()
{

}

void MiniAudioEngine::audioDataCallback(ma_device* pDevice, void* pOutput, const void* pInput, ma_uint32 frameCount)
{
    (void)pInput;
    //Do master audio processing before sending to device.
    ma_engine_read_pcm_frames((ma_engine*)pDevice->pUserData, pOutput, frameCount, NULL);
}

void MiniAudioEngine::stopEngine()
{
    ma_engine_uninit(&engine);
    ma_device_uninit(&device);
    ma_context_uninit(&context);
    ma_resource_manager_uninit(&resourceManager);
}

bool MiniAudioEngine::startEngine(int n)
{
    ma_result result;

    result = this->startContext();
    if (result != MA_SUCCESS) {
        return result;
    }
    this->getAllAudioDevices();
    result = this->startDevice(n);
    return result;
}

ma_result MiniAudioEngine::startDevice(int id)
{
    ma_result result;
    ma_device_config deviceConfig;
    ma_engine_config engineConfig;

    deviceConfig = ma_device_config_init(ma_device_type_playback);
    deviceConfig.playback.pDeviceID = &pPlaybackDeviceInfos[id].id;
    deviceConfig.playback.format    = resourceManager.config.decodedFormat;
    deviceConfig.playback.channels  = 0;
    deviceConfig.sampleRate         = resourceManager.config.decodedSampleRate;
    deviceConfig.dataCallback       = audioDataCallback;
    deviceConfig.pUserData          = &engine;
    result = ma_device_init(&context, &deviceConfig, &device);
    if (result != MA_SUCCESS) {
        printf("Failed to initialize audio device %s.", pPlaybackDeviceInfos[id].name);
        return result;
    }
    engineConfig = ma_engine_config_init();
    engineConfig.pDevice          = &device;
    engineConfig.pResourceManager = &resourceManager;
    engineConfig.noAutoStart      = MA_TRUE;
    result = ma_engine_init(NULL, &engine);
    if (result != MA_SUCCESS) {
        printf("Failed to initialize audio engine.");
        return result;
    }
    result = ma_engine_start(&engine);
    if (result != MA_SUCCESS) {
        printf("Failed to start audio engine %i.", id);
        return result;
    }
    iChosenDevice = id;
    printf("Initialized audio device %d: %s", id, pPlaybackDeviceInfos[id].name);
    return result;
}

ma_result MiniAudioEngine::startContext()
{
    ma_result result;

    resourceManagerConfig = ma_resource_manager_config_init();
    resourceManagerConfig.decodedFormat     = ma_format_f32;    /* ma_format_f32 should almost always be used as that's what the engine (and most everything else) uses for mixing. */
    resourceManagerConfig.decodedChannels   = 0;
    resourceManagerConfig.decodedSampleRate = 0;
    resourceManagerConfig.jobThreadCount    = 0;
    result = ma_resource_manager_init(&resourceManagerConfig, &resourceManager);
    if (result != MA_SUCCESS) {
        printf("Failed to initialize audio resource manager.");
        return result;
    }
    result = ma_context_init(NULL, 0, NULL, &context);
    if (result != MA_SUCCESS) {
        printf("Failed to initialize audio context.");
    }
    return result;
}

// enum all audio devices in system
ma_result MiniAudioEngine::getAllAudioDevices()
{
    ma_result result;

    result = ma_context_get_devices(&context, &pPlaybackDeviceInfos, &playbackDeviceCount, NULL, NULL);
    if (result != MA_SUCCESS) {
        printf("Failed to enumerate playback devices.\n");
        ma_context_uninit(&context);
        return result;
    }
    printf("Audio devices detected in system:\n");
    for (ma_uint32 iAvailableDevice = 0; iAvailableDevice < playbackDeviceCount; iAvailableDevice += 1) {
        qInfo("%d: : %s", iAvailableDevice, pPlaybackDeviceInfos[iAvailableDevice].name);
    }
    return result;
}

ma_result MiniAudioEngine::loadMedia(int layer, char *file)
{
    ma_result result;

    if (m_mediaLoaded[layer] == true)
    {
        qInfo("removing sound %i", layer);
        ma_sound_uninit(&m_currentSound[layer]);
        m_mediaLoaded[layer] = false;
    }
    result = ma_sound_init_from_file(&engine, file, \
                MA_SOUND_FLAG_NO_SPATIALIZATION \
                /*| MA_RESOURCE_MANAGER_DATA_SOURCE_FLAG_DECODE \
                  | MA_RESOURCE_MANAGER_DATA_SOURCE_FLAG_ASYNC \
                  | MA_RESOURCE_MANAGER_DATA_SOURCE_FLAG_STREAM \*/
                , NULL, NULL, &m_currentSound[layer]);
    if (result != MA_SUCCESS)
        qWarning("Failed to load file %s", file);
    else {
        m_mediaLoaded[layer] = true;
        this->volChanged(layer, 0);
    }
    return result;
}

float MiniAudioEngine::getDuration(int layer)
{
    ma_result result;
    float ret = 0;

    if (m_mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    result = ma_sound_get_length_in_seconds(&m_currentSound[layer], &ret);
    if (result != MA_SUCCESS)
    {
        qWarning("Can not get duration %i", layer);
        ret = 0;
    }
    return ret;
}

float MiniAudioEngine::getCursor(int layer)
{
    ma_result result;
    float ret = 0;

    if (m_mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    result = ma_sound_get_cursor_in_seconds(&m_currentSound[layer], &ret);
    if (result != MA_SUCCESS)
    {
        qWarning("Can not get cursor %i", layer);
        ret = 0;
    }
    return ret;
}

ma_result MiniAudioEngine::printFormatInfo(int layer)
{
    ma_format format;
    ma_uint32 channels;
    ma_uint32 sampleRate;

    if (m_mediaLoaded[layer] == false)
        return MA_DOES_NOT_EXIST;
    ma_result result = ma_sound_get_data_format(&m_currentSound[layer], &format, &channels, &sampleRate, NULL, 0);
    if (result != MA_SUCCESS) {
            qWarning("Failed to get data format %i\n", layer);
            return MA_INVALID_DATA;
    }
    qInfo("samples/sec: %u format: %u channels: %u", sampleRate, format, channels);
    return result;
}

// Expects between 0 and 100 vol value in db
void MiniAudioEngine::volChanged(int layer, float vol)
{
    float result;

    if (m_mediaLoaded[layer] == false)
        return;
    result = ma_volume_linear_to_db(1.00000000 + (vol / 800.0));
    ma_sound_group_set_volume(&m_currentSound[layer], result);
}

void MiniAudioEngine::panChanged(int layer, float value)
{
    float result;

    if (m_mediaLoaded[layer] == false)
        return;
    result = (value / 128.0) - 1.0;
    ma_sound_group_set_pan(&m_currentSound[layer], result);
}

void MiniAudioEngine::pitchChanged(int layer, float value)
{
    float result;

    if (m_mediaLoaded[layer] == false)
        return;
    result = value / 128.0;
    ma_sound_group_set_pitch(&m_currentSound[layer], result);
}

void MiniAudioEngine::playbackChanged(int layer, Status status)
{
    if (m_mediaLoaded[layer] == false)
        return;
    switch (status) {
        case Status::Paused:
            ma_sound_stop(&m_currentSound[layer]);
            break;
        case Status::Stopped:
            ma_sound_stop(&m_currentSound[layer]);
            ma_sound_seek_to_pcm_frame(&m_currentSound[layer], 0);
            break;
        case Status::PlayingLoop:
            ma_sound_set_looping(&m_currentSound[layer], true);
            ma_sound_start(&m_currentSound[layer]);
        break;
        case Status::PlayingOnce:
            ma_sound_set_looping(&m_currentSound[layer], false);
            ma_sound_start(&m_currentSound[layer]);
            break;
    }
}
